#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#       combac.py
#
#       Copyright 2014 BFR <info@freie-radios.de>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; Version 3 of the License
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, the license can be downloaded here:
#
#       http://www.gnu.org/licenses/gpl.html
import codecs
import os
import sys
import threading
import simplejson
import queue
import traceback

from datetime import datetime, timedelta

from libraries.database.broadcasts import Schedule, ScheduleEntry
from modules.communication.redis.messenger import RedisMessenger


class AuraCalendarService(threading.Thread):
    messenger = RedisMessenger()
    calendarurl = ""
    audiobase = ""
    playlistdir = ""
    until = ''
    xmlplaylist = range(0)

    queue = None
    config = None
    debug = False
    _stop_event = None

    """
    Fetching playlist data, write it into the database and notify service
    """
    def __init__(self, config, datefrom="", dateto="", debug=True):
        threading.Thread.__init__(self)
        self.config = config
        self.debug = debug

        self.messenger.set_channel('aura')
        self.messenger.set_section('calendar')
        self.messenger.set_mail_addresses(self.config.get('frommail'), self.config.get('adminmail'))

        self.dateto = dateto
        self.datefrom = str(datefrom)
        self.queue = queue.Queue()

        self._stop_event = threading.Event()

    # ------------------------------------------------------------------------------------------ #
    def set_date_from(self, date):
        self.datefrom = str(date).replace(" ", "T")

    # ------------------------------------------------------------------------------------------ #
    def set_date_to(self, date):
        self.dateto = str(date).replace(" ", "T")

    # ------------------------------------------------------------------------------------------ #
    def set_until_time(self, timestring):
        self.until = timestring

    # ------------------------------------------------------------------------------------------ #
    def set_calendar_url(self, url):
        self.calendarurl = url

    # ------------------------------------------------------------------------------------------ #
    def set_audio_path(self, path):
        self.audiobase = path

    # ------------------------------------------------------------------------------------------ #
    def set_playlist_store(self, path):
        self.playlistdir = path

    # ------------------------------------------------------------------------------------------ #
    def get_duration(self, start, end):
        return self.__calc_duration__(start, end)

    # ------------------------------------------------------------------------------------------ #
    def get_queue(self):
        return self.queue

    # ------------------------------------------------------------------------------------------ #
    def get_uri(self):
        if not self.playlistdir:
            return False
        if not self.datefrom:
            return False
        if not self.__calc_date_to__():
            return

        hostname = self.get('servername');
        port = self.get('serviceport');
        date_from = self.datefrom[0:16] + ':00';
        date_to = self.dateto[0:16] + ':00';
        uri = 'http://' + hostname  + ':' + port + '/playlist/' + date_from + '/' + date_to

        return uri

    # ------------------------------------------------------------------------------------------ #
    def run(self):
        """
        Fetch calendar data and store it in the database
        """

        self.__set_url__()
        # fetch upcoming schedules from ENGINE
        self.__fetch_schedule_data__()
        # fetch playlist and fallbacks to the schedules from TANK
        self.__fetch_schedule_entry_data__()

        for schedule in self.fetched_schedule_data:
            if 'schedule_start' not in schedule:
                print("WARNING: no start of schedule given. skipping the schedule")
                continue
            if 'schedule_end' not in schedule:
                print("WARNING: no end of schedule given. skipping the schedule")
                continue

            # store the schedule
            self.store_schedule(schedule)

            # store playlists to play
            entrynum = 0
            for entry in schedule["playlist"]["entries"]:
                self.store_schedule_entry(schedule, entry, entrynum)
                entrynum = entrynum + 1

        # release the mutex
        self.queue.put("fetching_finished")

        # terminate the thread
        return

    def store_schedule(self, schedule):
        from libraries.database.database import db
        #schedule_db = Schedule.query.filter(Schedule.schedule_id == schedule['schedule_id']).first()
        schedule_db = db.session.query(Schedule).filter(Schedule.schedule_id == schedule['schedule_id']).first()
#        schedule_db = Schedule.get_by_id(schedule['schedule_id'])

        if not schedule_db:
            print("no schedule with given schedule id in database => create new")
            schedule_db = Schedule()

        # calc duration
        duration = self.__calc_duration__(schedule['schedule_start'], schedule['schedule_end'])
        schedule['duration'] = timedelta(seconds=duration).__str__()

        schedule_db.show_id = schedule["show_id"]
        schedule_db.schedule_id = schedule["schedule_id"]
        schedule_db.schedule_start = schedule["schedule_start"]
        schedule_db.schedule_end = schedule["schedule_end"]
        schedule_db.show_name = schedule["show_name"]
        schedule_db.show_hosts = schedule["show_hosts"]
        schedule_db.is_repetition = schedule["is_repetition"]
        schedule_db.rtr_category = schedule["rtr_category"]
        schedule_db.comment = schedule["comment"]
        schedule_db.languages = schedule["languages"]
        schedule_db.type = schedule["type"]
        schedule_db.category = schedule["category"]
        schedule_db.topic = schedule["topic"]
        schedule_db.musicfocus = schedule["musicfocus"]

        schedule_db.playlist_id = schedule["playlist_id"]
        schedule_db.schedule_fallback_id = schedule["schedule_fallback_id"]
        schedule_db.show_fallback_id = schedule["show_fallback_id"]
        schedule_db.station_fallback_id = schedule["station_fallback_id"]

        schedule_db.store(True)

        event = {'job': 'dump', 'location': schedule["schedule_id"], 'length': duration}
        self.messenger.queue_add_event('dumpstart', schedule["schedule_start"], event)
        event = {'job': 'dump', 'location': schedule["schedule_id"], 'length': duration}
        self.messenger.queue_add_event('dumpend', schedule["schedule_end"], event)

        return schedule_db

    # ------------------------------------------------------------------------------------------ #
    def store_schedule_entry(self, schedule, entry, entrynum):
        schedule_entry = ScheduleEntry.select_one(schedule["playlist_id"], entrynum)

        if not schedule_entry:
            if self.debug:
                print("no scheduleentry with id " + str(schedule["playlist_id"]) + " and pos " + str(entrynum) + " in database => creating a new one")
            schedule_entry = ScheduleEntry()

        schedule_entry.playlist_id = schedule["playlist_id"]
        schedule_entry.entry_num = entrynum
        schedule_entry.schedule_id = schedule["schedule_id"]
        schedule_entry.entry_start = entry["entry_start"]
        schedule_entry.entry_end = entry["entry_end"]
        schedule_entry.source = entry["source"]
        schedule_entry.artist = entry["artist"]
        schedule_entry.track = entry["track"]
        schedule_entry.albumname = entry["albumname"]
        schedule_entry.genre = entry["genre"]
        schedule_entry.tracknum = entry["tracknum"]
        schedule_entry.cdnum = entry["cdnum"]
        schedule_entry.year = entry["year"]
        if schedule_entry.year == '':
            schedule_entry.year = 0

        if self.debug:
            print("Storing entries... playlist_id: " + str(schedule["playlist_id"]) + " schedule_id: " + str(
                schedule["schedule_id"]) + " num: " + str(entrynum))

        schedule_entry.store(True)

        return schedule_entry

    # ------------------------------------------------------------------------------------------ #
    def __calc_date_to__(self):
        if self.dateto:
            return True
        if not self.until:
            return False
        if not self.datefrom:
            return False

        date_start = datetime.strptime(self.datefrom.replace('T',' '), "%Y-%m-%d %H:%M")
        time_start = date_start.strftime('%H:%M')
        day_offset = 1 if (time_start > self.until) else 0
        end_date = date_start + timedelta(day_offset)
        self.dateto = end_date.strftime('%F') + 'T' + self.until

        return True

    # ------------------------------------------------------------------------------------------ #
    @staticmethod
    def __calc_duration__(start, end):
        """
        Berechnet Zeit in Sekunden aus Differenz zwischen Start und Enddatum
        @type start: datetime
        @param start: Startzeit
        @type end: datetime
        @param end: Endzeit
        @rtype:       int
        @return:      Zeit in Sekunden
        """
        sec1 = int(datetime.strptime(start[0:16].replace(" ","T"),"%Y-%m-%dT%H:%M").strftime("%s"));
        sec2 = int(datetime.strptime(end[0:16].replace(" ","T"),"%Y-%m-%dT%H:%M").strftime("%s"));
        return (sec2 - sec1);

    # ------------------------------------------------------------------------------------------ #
    def __fetch_schedule_entry_data__(self):
        # store fetched entries => do not have to fetch playlist_id more than once
        fetched_entries=[]

        try:
            url = self.get_url()
            data = self.get_data()
        except IOError as e:
            self.messenger.send("Could not connect to service " + self.dataURL, '1101', 'error', 'fetchCalenderData', self.__get_error_data__(), 'getcalendar')

        try:
            for schedule in self.fetched_schedule_data:
                # retrieve playlist and the fallbacks for every schedule
                # if a playlist is already fetched, it is not fetched again
                schedule["playlist"]          = self.__fetch_schedule_entries__(schedule["schedule_id"], schedule["playlist_id"],          fetched_entries)
                schedule["schedule_fallback"] = self.__fetch_schedule_entries__(schedule["schedule_id"], schedule["schedule_fallback_id"], fetched_entries)
                schedule["show_fallback"]     = self.__fetch_schedule_entries__(schedule["schedule_id"], schedule["show_fallback_id"],     fetched_entries)
                schedule["station_fallback"]  = self.__fetch_schedule_entries__(schedule["schedule_id"], schedule["station_fallback_id"],  fetched_entries)

                if self.debug:
                    print("fetched for schedule_id: "+str(schedule["schedule_id"]))
                    print("playlist: "+str(schedule["playlist"]))
                    print("schedule_fallback: "+str(schedule["schedule_fallback"]))
                    print("show_fallback: "+str(schedule["show_fallback"]))
                    print("station_fallback: "+str(schedule["station_fallback"]))

        except Exception as e:
            traceback.print_exc()
            print("houston we have a problem")

    # ------------------------------------------------------------------------------------------ #
    def __fetch_schedule_entries__(self, schedule_id, playlist_id, fetched_schedule_entries):
        for entry in fetched_schedule_entries:
            if entry["playlist_id"] == playlist_id:
                if self.debug:
                    print(str(playlist_id) + " already fetched")
                return entry

        if self.debug:
            print("Wanna fetch playlist/pool id " + str(playlist_id) + " of schedule #" + str(schedule_id))

        # HARDCODED Testdata
        if playlist_id == 1:  # station fallback
            schedule_entries = simplejson.loads('{"playlist_id":1,"entries":[{"source":"pool://easylistening","artist":"station wide fallback pool","track":"","albumname":"","genre":"","tracknum":"","cdnum":"","year":""}]}')
        if playlist_id == 2:  # playlist to play for schedule #1 => linein
            schedule_entries = simplejson.loads('{"playlist_id":2,"entries":[{"entry_start":"'+(datetime.now()+timedelta(hours=0)).strftime('%Y-%m-%d %H:00:00')+'","entry_end":"'+(datetime.now()+timedelta(hours=0)).strftime('%Y-%m-%d %H:05:00')+'","source":"/var/audio/jingles/jingle.flac","artist":"jingle","track":"","albumname":"","genre":"","tracknum":"","cdnum":"","year":"1999"},{"entry_start":"'+(datetime.now()+timedelta(hours=0)).strftime('%Y-%m-%d %H:05:00')+'","entry_end":"'+(datetime.now()+timedelta(hours=1)).strftime('%Y-%m-%d %H:00:00')+'","source":"linein://0","artist":"live","track":"","albumname":"","genre":"","tracknum":"","cdnum":"","year":""}]}')
        if playlist_id == 4:  # playlist to play for schedule #2 => from file system
            schedule_entries = simplejson.loads('{"playlist_id":4,"entries":[{"entry_start":"'+(datetime.now()+timedelta(hours=1)).strftime('%Y-%m-%d %H:00:00')+'","entry_end":"'+(datetime.now()+timedelta(hours=1)).strftime('%Y-%m-%d %H:05:00')+'","source":"file:///var/audio/jingles/jingle.flac","artist":"jingle","track":"","albumname":"","genre":"","tracknum":"","cdnum":"","year":"1999"},{"entry_start":"'+(datetime.now()+timedelta(hours=1)).strftime('%Y-%m-%d %H:05:00')+'","entry_end":"'+(datetime.now()+timedelta(hours=2)).strftime('%Y-%m-%d %H:00:00')+'","source":"file:///var/audio/preprod.flac","artist":"live","track":"","albumname":"","genre":"","tracknum":"","cdnum":"","year":""}]}')
        if playlist_id == 6:  # playlist to play for schedule #3 => http stream
            schedule_entries = simplejson.loads('{"playlist_id":6,"entries":[{"entry_start":"'+(datetime.now()+timedelta(hours=2)).strftime('%Y-%m-%d %H:00:00')+'","entry_end":"'+(datetime.now()+timedelta(hours=3)).strftime('%Y-%m-%d %H:00:00')+'","source":"http://stream.fro.at:80/fro-128.ogg","artist":"Stream Übernahme von Radio FRO","track":"","albumname":"","genre":"","tracknum":"","cdnum":"","year":""}]}')
        if playlist_id == 12:  # schedule #0 fallback
            schedule_entries = simplejson.loads('{"playlist_id":12,"entries":[{"source":"pool:///drumandbass"}]}')
        if playlist_id == 22:  # schedule #1 fallback
            schedule_entries = simplejson.loads('{"playlist_id":22,"entries":[{"source":"pool:///chillout"}]}')
        if playlist_id == 32:  # schedule #2 fallback
            schedule_entries = simplejson.loads('{"playlist_id":22,"entries":[{"source":"pool:///rock"}]}')
        if playlist_id == 92:  # show fallback for schedule #0
            schedule_entries = simplejson.loads('{"playlist_id":92,"entries":[{"source":"file:///var/audio/playlist/playlist1.m3u"}]}')
        if playlist_id == 102:  # show fallback for schedule #1
            schedule_entries = simplejson.loads('{"playlist_id":102,"entries":[{"source":"file:///var/audio/playlist/playlist2.m3u"}]}')
        if playlist_id == 112:  # show fallback for schedule #2
            schedule_entries = simplejson.loads('{"playlist_id":112,"entries":[{"source":"file:///var/audio/playlist/playlist3.m3u"}]}')

        for entry in schedule_entries["entries"]:
            if entry["source"].startswith("file"):
                if not os.path.isfile(entry["source"]):
                    print("WARNING: File", entry["source"], "does not exist!")
#            elif entry["source"].startswith("pool"):

        fetched_schedule_entries.append(schedule_entries)

        return schedule_entries

    # ------------------------------------------------------------------------------------------ #
    def __fetch_schedule_data__(self):
        # Now open Calendar Url
        try:
            url = self.get_url()
            data = self.get_data()

            #request = urllib.request.Request(url, data)
            #with urllib.request.urlopen(request) as response:
            #    html_response = response.read()

            html_response = ''

        except IOError as e:
            self.messenger.send("Could not connect to service " + self.dataURL, '1101', 'error', 'fetchCalenderData',
                                self.__get_error_data__(), 'getcalendar')
            #TODO: so gehts nicht
            sys.exit()
        else:
            # Read in data
            #print(html_response[0:100])
            #jsondata = html_response # .read()

            jsonstring = '[{"schedule_id":1,"schedule_start":"' + (datetime.now() + timedelta(hours=0)).strftime('%Y-%m-%d %H:00:00') + '","schedule_end":"' + (datetime.now() + timedelta(hours=1)).strftime('%Y-%m-%d %H:00:00') + '","show_id":9,"show_name":"FROzine","show_hosts":"Sandra Hochholzer, Martina Schweiger","is_repetition":false,"playlist_id":2,"schedule_fallback_id":12,"show_fallback_id":92,"station_fallback_id":1,"rtr_category":"string","comment":"Kommentar","languages":"Sprachen","type":"Typ","category":"Kategorie","topic":"Topic","musicfocus":"Fokus"},{"schedule_id":2,"schedule_start":"' + (datetime.now()+timedelta(hours=1)).strftime('%Y-%m-%d %H:00:00') + '","schedule_end":"' + (datetime.now()+timedelta(hours=2)).strftime('%Y-%m-%d %H:00:00') + '","show_id":10,"show_name":"FROMat","show_hosts":"Sandra Hochholzer, Martina Schweiger","is_repetition":false,"playlist_id":4,"schedule_fallback_id":22,"show_fallback_id":102,"station_fallback_id":1,"rtr_category":"string","comment":"Kommentar","languages":"Sprachen","type":"Typ","category":"Kategorie","topic":"Topic","musicfocus":"Fokus"},{"schedule_id":3,"schedule_start":"' + (datetime.now()+timedelta(hours=2)).strftime('%Y-%m-%d %H:00:00') + '","schedule_end":"' + (datetime.now() + timedelta(hours=3)).strftime('%Y-%m-%d %H:00:00') + '","show_id":11,"show_name":"Radio für Senioren","show_hosts":"Sandra Hochholzer, Martina Schweiger","is_repetition":false,"playlist_id":6,"schedule_fallback_id":32,"show_fallback_id":112,"station_fallback_id":1,"rtr_category":"string","comment":"Kommentar","languages":"Sprachen","type":"Typ","category":"Kategorie","topic":"Topic","musicfocus":"Fokus"}]'

            try:
                self.fetched_schedule_data = simplejson.loads(jsonstring)
            except Exception as e:
                self.messenger.send("Could not decode calender data from service " + self.dataURL + "! Exception: " + e, '1102', 'error',
                                    'fetchCalenderData', self.__get_error_data__(), 'getcalendar')
                sys.exit()
            else:
                # check data
                try:
                    print("WARNING: Hardcoded Response && no json data checks. i believe what i get here")
                    #print(self.data)
                    #print(len(self.data))
                except KeyError as e:
                    self.messenger.send("Could not decode calender data from service " + self.dataURL + "! Exception: " + e, '1102', 'error',
                                        'fetchCalenderData', self.__get_error_data__(), 'getcalendar')
                    sys.exit()
                except Exception as e:
                    self.messenger.send("Could not decode calender data from service " + self.dataURL + "! Exception: " + e, '1102', 'error',
                                        'fetchCalenderData', self.__get_error_data__(), 'getcalendar')
                    sys.exit()

                else:
                    return self.fetched_schedule_data

    # ------------------------------------------------------------------------------------------ #
    def __get_error_data__(self):
        """
        Basisdaten als dict liefern
        """
        return {'from': str(self.datefrom), 'dateto': str(self.dateto), 'path': "self.playlistpath", 'url': self.calendarurl}

    # ------------------------------------------------------------------------------------------ #
    def __set_url__(self):
        """
        Die URL auf den Zappa-Kalender ermitteln
        """
        self.dataURL = self.calendarurl.replace('#datefrom#', self.datefrom.replace(' ', 'T')).replace('#dateto#', self.dateto.replace(' ', 'T'))

    # ------------------------------------------------------------------------------------------ #
    def get_url(self):
        url = self.dataURL[0:self.dataURL.find('&')]
        return url

    # ------------------------------------------------------------------------------------------ #
    def get_data(self):
        params = self.dataURL[self.dataURL.find('&'):]
        print("params: "+params)

    # ------------------------------------------------------------------------------------------ #
    def stop(self):
        self._stop_event.set()

    # ------------------------------------------------------------------------------------------ #
    def get_calendar_data(self):
        return self.fetched_schedule_data
